// 32feet.NET - Personal Area Networking for .NET
//
// InTheHand.Net.Bluetooth.Widcomm.WidcommBluetoothFactoryBase
// 
// Copyright (c) 2008-2010 In The Hand Ltd, All rights reserved.
// This source code is licensed under the In The Hand Community License - see License.txt

using System;
using System.Collections.Generic;
using System.Text;
using InTheHand.Net.Sockets;
using System.Diagnostics;
using InTheHand.Net.Bluetooth.Factory;

namespace InTheHand.Net.Bluetooth.Widcomm
{
    abstract class WidcommBluetoothFactoryBase : BluetoothFactory
    {
        internal abstract WidcommBtInterface GetWidcommBtInterface();
        internal abstract WidcommRfcommStreamBase GetWidcommRfcommStream();
        internal abstract WidcommRfcommStreamBase GetWidcommRfcommStreamWithoutRfcommIf();
        internal abstract IRfcommPort GetWidcommRfcommPort();
        internal abstract IRfCommIf GetWidcommRfCommIf();
        internal abstract ISdpService GetWidcommSdpService();
        internal abstract void EnsureLoaded();
        //
#if WIDCOMM_SINGLE_THREADING || WinXP
        internal abstract WidcommPortSingleThreader GetSingleThreader();
#endif

        #region Resource Cleanup records
        internal void AddPort(WidcommRfcommStreamBase port)
        {
            lock (_livePorts) {
                _livePorts.Add(port);
            }
        }

        internal void RemovePort(WidcommRfcommStreamBase port)
        {
            bool found;
            lock (_livePorts) {
                found = _livePorts.Remove(port);
            }
            Debug.WriteLine(DateTime.Now.TimeOfDay.ToString()
                + ": RemovePort found it: " + found);
        }

        internal WidcommRfcommStreamBase[] GetPortList()
        {
            lock (_livePorts) {
                var a = _livePorts.ToArray();
                Debug.WriteLine(DateTime.Now.TimeOfDay.ToString() + ": GetPortList returned " + a.Length + " ports.");
                return a;
            }
        }

        List<WidcommRfcommStreamBase> _livePorts = new List<WidcommRfcommStreamBase>();
        #endregion
    }

    sealed class WidcommBluetoothFactory : WidcommBluetoothFactoryBase
    {
        static IBtIf s_btIf;
        static WidcommBtInterface s_btInterface;
        static WidcommBluetoothSecurity m_btSecurity;
#if WIDCOMM_SINGLE_THREADING || WinXP
        static WidcommPortSingleThreader _st;
#endif
        internal volatile bool _seenStackDownEvent;

        public WidcommBluetoothFactory()
        {
            EnsureLoaded();
            // Check radio is connected.
            this.GetPrimaryRadio();
        }

        internal override void EnsureLoaded()
        {
            lock (typeof(WidcommBluetoothFactory)) {
                // In case Widcomm is forced to shutdown when CBtIf is extant we monitor
                // for the stack-down event so we can force reload next time
                // we're used.
                if (_seenStackDownEvent) {
                    WidcommUtils.Trace_WriteLine("Restarting due to stack/radio shutdown.");
                    Dispose(true);
                    Debug.Assert(s_btIf == null, "After Dispose but NOT s_btIf == null");
                    Debug.Assert(s_btInterface == null, "After Dispose but NOT s_btInterface == null");
                }
                _seenStackDownEvent = false;
                //----
                // CBtIf: MUST ONLY be ONE of these, and must be FIRST created!
                //   "Only one object of this class should be instantiated by the application."
                //   "This class must be instantiated before any other DK classes are used"
                if (s_btIf == null) {
                    IBtIf btIf = new WidcommBtIf(this);
#if WIDCOMM_SINGLE_THREADING || WinXP
#if KILL_SINGLE_THREAD_AT_DISPOSAL
                    Debug.Assert(_st == null);
#endif
                    if (_st == null) _st = new WidcommPortSingleThreader();
                    Debug.Assert(GetSingleThreader() != null);
                    if (GetSingleThreader() != null) {
                        btIf = new WidcommStBtIf(this, btIf);
                        WidcommUtils.Trace_WriteLine("IBtIf using WidcommStBtIf.");
                    }
#endif
                    Debug.Assert(s_btInterface == null);
                    WidcommBtInterface btInterface = new WidcommBtInterface(btIf, this);
                    // Don't set these until we're sure that initialisation has 
                    // all completed successfully.
                    s_btIf = btIf;
                    s_btInterface = btInterface;
                } else {
                    Debug.Assert(s_btInterface != null, "One set but not the other!");
                }
            }//lock
        }

        //-----
        protected override void Dispose(bool disposing)
        {
            IDisposable iface;
            IDisposable st;
            lock (typeof(WidcommBluetoothFactory)) {
                iface = (IDisposable)s_btInterface;
                s_btIf = null;
                s_btInterface = null;
                m_btSecurity = null;
#if (WIDCOMM_SINGLE_THREADING || WinXP) && KILL_SINGLE_THREAD_AT_DISPOSAL
                st = _st;
                _st = null;
#else
                st = null;
#endif
            }//lock
            if (iface != null)
                iface.Dispose();
            if (st != null)  // Must NOT dispose this first!
                st.Dispose();
        }

        //-----
#if WIDCOMM_SINGLE_THREADING || WinXP
        internal override WidcommPortSingleThreader GetSingleThreader()
        {
            return _st;
        }
#endif

        //-----
        protected override IBluetoothClient GetBluetoothClient()
        {
            EnsureLoaded();
            return new WidcommBluetoothClient(this);
        }

        protected override IBluetoothClient GetBluetoothClient(System.Net.Sockets.Socket acceptedSocket)
        {
            throw new NotSupportedException("Cannot create a BluetoothClient from a Socket on the Widcomm stack.");
        }

        protected override IBluetoothClient GetBluetoothClientForListener(CommonRfcommStream strm)
        {
            return new WidcommBluetoothClient((WidcommRfcommStreamBase)strm, this);
        }

        protected override IBluetoothClient GetBluetoothClient(BluetoothEndPoint localEP)
        {
            EnsureLoaded();
            return new WidcommBluetoothClient(localEP, this);
        }

        //----------------
        protected override IBluetoothListener GetBluetoothListener()
        {
            EnsureLoaded();
            return new WidcommBluetoothListener(this);
        }

        //----------------
        internal override WidcommBtInterface GetWidcommBtInterface()
        {
            EnsureLoaded();
            return s_btInterface;
        }

        internal override WidcommRfcommStreamBase GetWidcommRfcommStream()
        {
            EnsureLoaded();
            return new WidcommRfcommStream(GetWidcommRfcommPort(), GetWidcommRfCommIf(), this);
        }

        internal override WidcommRfcommStreamBase GetWidcommRfcommStreamWithoutRfcommIf()
        {
            EnsureLoaded();
            return new WidcommRfcommStream(GetWidcommRfcommPort(), null, this);
        }

        internal override IRfcommPort GetWidcommRfcommPort()
        {
            EnsureLoaded();
            // Handling of single threadedness is done within WidcommRfcommStream.
            return new WidcommRfcommPort();
        }

        internal override IRfCommIf GetWidcommRfCommIf()
        {
            EnsureLoaded();
            IRfCommIf inst = new WidcommRfCommIf();
#if WIDCOMM_SINGLE_THREADING || WinXP
            if (GetSingleThreader() != null) {
                inst = new WidcommStRfCommIf(this, inst);
                WidcommUtils.Trace_WriteLine("IRfCommIf using WidcommStRfCommIf.");
            }
#endif
            return inst;
        }

        protected override IBluetoothDeviceInfo GetBluetoothDeviceInfo(BluetoothAddress address)
        {
            EnsureLoaded();
            return WidcommBluetoothDeviceInfo.CreateFromGivenAddress(address, this);
        }

        //----------------
        protected override IBluetoothRadio GetPrimaryRadio()
        {
            EnsureLoaded();
            return new WidcommBluetoothRadio(this);
        }

        protected override IBluetoothRadio[] GetAllRadios()
        {
            EnsureLoaded();
            // Widcomm supports only one radio.
            return new IBluetoothRadio[] { GetPrimaryRadio() };
        }

        //----------------
        internal override ISdpService GetWidcommSdpService()
        {
            EnsureLoaded();
            return new SdpService();
        }

        //----------------
        protected override IBluetoothSecurity GetBluetoothSecurity()
        {
            EnsureLoaded();
            if (m_btSecurity == null) {
                m_btSecurity = new WidcommBluetoothSecurity(this);
            }
            return m_btSecurity;
        }

    }
}
