using System;
using System.Collections.Generic;
using System.Text;
using System.Threading;

namespace RobotCtrl
{
    /// <summary>
    /// Treiber-SW zum Motor-Controller-IC LM629
    /// </summary>
    public class MotorCtrl_HW : MotorCtrl
    {
        protected const double samplePeriod = 256E-6;
        protected const double speedScale = samplePeriod / Config.MeterPerTick * (1 << 16);
        protected const double accelerationScale = samplePeriod * samplePeriod / Config.MeterPerTick * (1 << 16);

        private readonly int io;
        private readonly object hwLock = new object();

        public MotorCtrl_HW(int IOAddress)
        {
            io = IOAddress;
            reset();
        }


        public override int Status { get { return readStatus(); } }
        public override bool Ready { get { return (readStatus() & 0x01) == 0; } }

        public override double Acceleration // m/(s*s)
        {
            set
            {
                lock (hwLock)
                {
                    int accel = (int)(accelerationScale * value);
                    writeCmd(0x1F);
                    writeShort(0x0820);
                    writeInt(accel);
                }
            }
        }

        public override double Speed // m/s
        {
            set
            {
                lock (hwLock)
                {
                    int velocity = (int)Math.Abs(speedScale * value);
                    int mode = (value >= 0) ? 0x1808 : 0x0808;
                    writeCmd(0x1F);
                    writeShort(mode);
                    writeInt(velocity);
                    nominalSpeed = value;
                }
            }
            get { return nominalSpeed; }
        }

        protected override int Ticks
        {
            get
            {
                lock (hwLock)
                {
                    writeCmd(0x0A);
                    return readInt();
                }
            }
        }

        public override void Close()
        {
            lock (hwLock)
            {
                stop();
            }
        }

        public override void Reset()
        {
            lock (hwLock)
            {
                reset();
            }
        }

        public override void Go()
        {
            lock (hwLock)
            {
                writeCmd(0x01);
            }
        }

        public override void Stop()
        {
            lock (hwLock)
            {
                stop();
            }
        }

        public override void SetPID(int proportional, int integral, int derivative, int integralLimit, int derivativeInterval)
        {
            lock (hwLock)
            {
                writeCmd(0x1E);
                writeShort(((derivativeInterval & 0xFF) << 8) | 0x0F);
                writeShort(proportional);
                writeShort(integral);
                writeShort(derivative);
                writeShort(integralLimit);
                writeCmd(0x04);
            }
        }

        void reset()
        {
            Thread.Sleep(20);
            IOPort.Write(io, 0x00);
            Thread.Sleep(20);
            IOPort.Write(io, 0x02);			// Define Home
            Thread.Sleep(50);
            //	SetPID(100, 20, 1000, 1000, 1);
            SetPID(1500, 200, 8000, 1000, 1);
        }

        void stop()
        {
            writeCmd(0x1F);
            writeShort(0x0100);
            writeCmd(0x01);
        }

        void waitReady()
        {
            waitReady(200);
        }

        void waitReady(int timeout)
        {
            const int sleepTime = 1;
            int time = 0;
            while (!Ready)
            {
                Thread.Sleep(sleepTime);
                time += sleepTime;
                if (time > timeout)
                    throw new ApplicationException("waitReady-Timeout: Motor Controller IC LM629.");
            }
        }

        void writeCmd(byte cmd)
        {
            waitReady();
            IOPort.Write(io, cmd);
        }

        void writeShort(int val)
        {
            waitReady();
            IOPort.Write(io + 1, (val >> 8) & 0xFF);
            IOPort.Write(io + 1, val & 0xFF);
        }

        void writeInt(int val)
        {
            writeShort(((val >> 16) & 0xFFFF));
            writeShort(((val >> 0) & 0xFFFF));
        }

        int readStatus()
        {
            return IOPort.Read(io);
        }

        int readShort()
        {
            waitReady();
            int val = IOPort.Read(io + 1);
            val = (val << 8) | IOPort.Read(io + 1);
            return val;
        }

        int readInt()
        {
            int val = readShort();
            val = (val << 16) | readShort();
            return val;
        }
    }
}
